#include "RainDrop256256.h"  

#include "RainDrop256256_coreE.h"
#include "RainDrop256256_coreD.h"

static void  EncRainRrop256256(u8i *out, u8i *cp, u8i *K);
static void RainDropGenRK_256_256(u8i * Rk, const u8i * K,int dirt);
static void RainDrop256256EncDataS_Round(u8i *out, u8i *cp, u8i *K,int round);

/*
only for full round enc & dec
enc roundkey size / dec roundkey size different
*/
void Key_Schedule(unsigned char *Seedkey, int KeyLen, unsigned   char Direction, unsigned char *Subkey)
{
	if (Direction == Dirt_ENC)
		RainDropGenRK_256_256_ENC(Subkey, Seedkey);
	else
		RainDropGenRK_256_256M_DEC((__m256i *)Subkey, Seedkey); 
}
int Crypt_Enc_Block_Round(unsigned char *input,int in_len,  unsigned char *output, int *out_len, unsigned char *key, int keylen, int CryptRound)
{
	u8i _C_ALIGN16( rk[RoundRainDrop256256*16]); 
	int i;
	if(in_len<1)
		return -1;
	if(keylen!=32)
		return -2;
	if((CryptRound> RoundRainDrop256256) || (CryptRound<0))
		return -3;  
	RainDropGenRK_256_256(rk, key,Dirt_ENC); 
	for(i=0;i<in_len;i+=32)
		RainDrop256256EncDataS_Round(output+i, input+i, rk, CryptRound);
	printf(" round   %03d : ",CryptRound);
	for(i=0;i<32;i++)
	{
		printf("%02X", output[i]);
		if(i==7 || i==23) 
			printf(" ");
		else if(i==15)
			printf("  ");
	}
	printf(" rk%03d : ",CryptRound);
	for(i=0;i<16;i++)
	{
      printf("%02X", rk[CryptRound*16+i-16]);
	  if(i==7)
		  printf(" ");
	}
	  printf("\n");
	*out_len= in_len;
	return 0; 
} 

int Crypt_Enc_Block(unsigned char *input,int in_len, unsigned char  *output, int *out_len, unsigned char *key , int keylen)
{
	u8i _C_ALIGN16( rk[RoundRainDrop256256*16]); 
	int i;
	if(in_len<1)
		return -1;
	if(keylen!=32)
		return -2; 
	RainDropGenRK_256_256_ENC(rk, key);
	for(i=0;i<in_len;i+=32)
		RainDrop256256EncDataS(output+i, input+i, rk);

	*out_len= in_len;
	return 0;
	 


}
 int Crypt_Dec_Block(unsigned char *input,int in_len, unsigned char  *output, int *out_len, unsigned char *key , int keylen)
 { 

	u8i TText[256];
	__m256i rksimd[RoundRainDrop256256 * 4];
	int leftbytes;
	if (in_len < 1)
		return -1;
	if (in_len & 0x0f)
		return -2;
	if (keylen != 32)
		return -3;
	RainDropGenRK_256_256M_DEC(rksimd, key);
	leftbytes = in_len;
	while (leftbytes >= 256)
	{
		RainDrop256256EncDataM(output, input, (u8i *)rksimd);
		output += 256;
		input += 256;
		leftbytes -= 256;
	}
	if (leftbytes)
	{
		memcpy(TText, input, leftbytes);
		RainDrop256256EncDataM(TText, TText, (u8i *)rksimd);
		memcpy(output, TText, leftbytes);
	}
	*out_len = in_len;
	return 0; 
 }

 //IV all 0

 int Crypt_Enc_Block_CBC(unsigned char *input, int in_len, unsigned char  *output, int *out_len, unsigned char *key, int keylen)
 {
	 u8i _C_ALIGN16(rk[RoundRainDrop256256 * 16]);
	 __m256i IV;
	 int i;
	 if (in_len < 1)
		 return -1;
	 if (in_len & 0x0f)
		 return -2;
	 if (keylen != 32)
		 return -3;
	 RainDropGenRK_256_256_ENC((u8i *)rk, key);
	 RainDrop256256EncDataS(output, input, (u8i *)rk);
	 for (i = 32; i < in_len; i += 32)
	 {
		 IV = _mm256_xor_si256(_mm256_loadu_si256( (__m256i *)(input+i)), _mm256_loadu_si256((__m256i *)(output + i-32)) );
		 RainDrop256256EncDataS(output + i, (u8i *)( &IV), (u8i *)rk);
	 }
	 *out_len = in_len;
	 return 0;
 }
 int Crypt_Dec_Block_CBC(unsigned char *input, int in_len, unsigned char  *output, int *out_len, unsigned char *key, int keylen)
 {
	 u8i _C_ALIGN16(TText[256]);
	 __m256i rksimd[RoundRainDrop256256 * 4];
	 __m256i mIV = _mm256_set1_epi64x(0);
	 int  leftbytes;
	 if (in_len < 1)
		 return -1;
	 if (in_len & 0x0f)
		 return -2;
	 if (keylen != 32)
		 return -3;
	 RainDropGenRK_256_256M_DEC(rksimd, key);
	 leftbytes = in_len;
	 while (leftbytes >= 256)
	 {
		 RainDrop256256EncDataM_CBC(output, input, (u8i *)rksimd, &mIV);
		 output += 256;
		 input += 256;
		 leftbytes -= 256;
	 }
	 if (leftbytes)
	 {
		 memcpy(TText, input, leftbytes);
		 RainDrop256256EncDataM_CBC(TText, TText, (u8i *)rksimd, &mIV);
		 memcpy(output, TText, leftbytes);
	 }
	 *out_len = in_len;
	 return 0;
 }

static void RainDropGenRK_256_256(u8i * Rk, const u8i * K,int dirt)
{ 
	u32i tmp;  	 
	u32i T0,T1,T2,T3,T4,T5,T6,T7;		
	u32i *p= (u32i*)Rk; 
	u32i i; 
	T0= SWAP32( *((u32i *) K));
	T1= SWAP32( *((u32i *) (K+4)));
	T2= SWAP32( *((u32i *) (K+8)));
	T3= SWAP32( *((u32i *) (K+12)));
	T4= SWAP32( *((u32i *) (K+16)));
	T5= SWAP32( *((u32i *) (K+20)));
	T6= SWAP32( *((u32i *) (K+24)));
	T7= SWAP32( *((u32i *) (K+28))); 
	if(dirt==  Dirt_ENC)
	{
		*p =  SWAP32(T0); 
		*(p+1) = SWAP32(T1);
		*(p+2) = SWAP32(T2);
		*(p+3) = SWAP32(T3);   
		p+=4; 
		for( i=1;i<RoundRainDrop256256;i++)  
		{ 	   
			T2^=i;  
			Fun_A(T,13,19,tmp);  
			Fun_A(T,13,19,tmp);   
			Fun_A(T,13,19,tmp);  
			Fun_A(T,13,19,tmp);     
			*p =  SWAP32(T0); 
			*(p+1) = SWAP32(T1);
			*(p+2) = SWAP32(T2);
			*(p+3) = SWAP32(T3);   
			p+=4; 
		}	
	}
	else
	{
		p+=RoundRainDrop256256*4-1;
		*p =  SWAP32(T3); 
		*(p-1) = SWAP32(T2);
		*(p-2) = SWAP32(T1);
		*(p-3) = SWAP32(T0);   
		p-=4; 
		for( i=1;i<RoundRainDrop256256;i++)  
		{ 	   
			T2^=i;  
			Fun_A(T,13,19,tmp);  
			Fun_A(T,13,19,tmp);   
			Fun_A(T,13,19,tmp);  
			Fun_A(T,13,19,tmp); 
			*p =  SWAP32(T3); 
			*(p-1) = SWAP32(T2);
			*(p-2) = SWAP32(T1);
			*(p-3) = SWAP32(T0);     
			p-=4; 
		}		
	}
}


static void SingleRoundRainDrop256256(u32i *in,u32i *rk)
{
	u32i l0,l1,l2,l3;
	u32i r0,r1,r2,r3;  
	u32i k0,k1,k2,k3;
	u32i t0,t1,t2,t3,t4,t5;
	l0=(*in);
	l1=*(in+1);
	l2=*(in+2);
	l3=*(in+3);
	r0=*(in);
	r1=*(in+1);
	r2=*(in+2);
	r3=*(in+3);    
	Sbox(l0);
	Sbox(l1);
	Sbox(l2);
	Sbox(l3); 

	k0= l1^l3; 
	k1= ROTL32(l2^l3,12);    
	k2= ROTL32(l0,14);
	k3= ROTL32(l0^l1,24);  

	l0=k0^(SWAP32(*(rk)) );
	l1=k1^(SWAP32(*(rk+1)) );
	l2=k2^(SWAP32(*(rk+2)) );
	l3=k3^(SWAP32(*(rk+3)) ); 
	(*in)=l0^(*(in+4));
	*(in+1)=l1^(*(in+5));
	*(in+2)=l2^(*(in+6));
	*(in+3)=l3^(*(in+7));   
	*(in+4)=r0;
	*(in+5)=r1;
	*(in+6)=r2;
	*(in+7)=r3; 
} 

static void  RainDrop256256EncDataS_Round(u8i *out, u8i *cp, u8i *rk,int round)
{ 
	int i;
	u32i sround256[8] ;
	sround256[0]= SWAP32( *((u32i *) cp));
	sround256[1]= SWAP32( *((u32i *) (cp+4)));
	sround256[2]= SWAP32( *((u32i *) (cp+8)));
	sround256[3]= SWAP32( *((u32i *) (cp+12)));
	sround256[4]= SWAP32(  *((u32i *) (cp+16)));
	sround256[5]= SWAP32( *((u32i *) (cp+20)));
	sround256[6]= SWAP32( *((u32i *) (cp+24)));
	sround256[7]= SWAP32( *((u32i *) (cp+28)));	 
	for(  i=0;i<round;i++)
	{ 
		SingleRoundRainDrop256256(sround256,(u32i *)(rk+i*16));
	}  
	*((u32i *) out)= SWAP32(sround256[4] );
	*((u32i *) (out+4))= SWAP32( sround256[5]);
	*((u32i *) (out+8))= SWAP32( sround256[6]  );
	*((u32i *) (out+12))= SWAP32( sround256[7]);   
	*((u32i *) (out+16))= SWAP32(sround256[0] );
	*((u32i *) (out+20))= SWAP32( sround256[1]);
	*((u32i *) (out+24))= SWAP32( sround256[2]  );
	*((u32i *) (out+28))= SWAP32( sround256[3]);    
}